/*
====================================================================================================

    Copyright (C) 2021 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.com/copyright-license

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

/* RENDERTARGETS: 6,7 */
layout(location = 0) out vec4 resultSH;
layout(location = 1) out vec4 resultAux;

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

in vec2 coord;

uniform sampler2D colortex6;
uniform sampler2D colortex7;
uniform sampler2D colortex9;
uniform sampler2D colortex10;

uniform float far;

uniform vec2 pixelSize, viewSize;

/* ------ ATROUS ------ */

const float gaussKernel[2][2] = {
	{ 1.0 / 4.0, 1.0 / 8.0  },
	{ 1.0 / 8.0, 1.0 / 16.0 }
};

float computeSigmaL(sampler2D img, ivec2 pixelUV, float center) {
	float sum = center * gaussKernel[0][0];

    const int r = 1;
	for(int y = -r; y <= r; y++) {
		for(int x = -r; x <= r; x++) {
			if(x != 0 || y != 0) {
				ivec2 p     = pixelUV + ivec2(x, y);
				float variance = texelFetch(img, p, 0).z;
				float w     = gaussKernel[abs(x)][abs(y)];
				sum        += variance * w;
			}
		}
	}

	return sqrt(max(sum, 1e-8));
}

#include "/lib/light/sh.glsl"

SH filterImage(sampler2D imgNormalDepth, out float varOut, vec2 uv, const int stepSize) {
	ivec2 pixelUV       = ivec2(uv * viewSize);

    vec4 normalDepthCenter  = texelFetch(imgNormalDepth, pixelUV, 0);

    vec4 colorCenterY   = texelFetch(colortex6, pixelUV, 0);
    vec4 colorCenterAux = texelFetch(colortex7, pixelUV, 0);

    SH colorCenter      = SH(colorCenterY, colorCenterAux.xy);

    if (normalDepthCenter.a == 1.0) return colorCenter;

        normalDepthCenter.rgb = normalDepthCenter.rgb * 2.0 - 1.0;
        normalDepthCenter.a = sqr(normalDepthCenter.a) * far;

	float lumaCenter = getLuma(SH_to_irradiance(colorCenter));

    float frames    = colorCenterAux.w * maxFrames;

    float normalExp = mix(2.0, 8.0, saturate(frames / 8.0)) * SVGF_NORMALEXP;

    float sigmaBias = (1.0 / max(frames, 1.0)) + 0.16;
    float maxDelta  = mix(0.71, tau, saturate(frames / 32.0));
    float offset    = mix(0.08, 0.02, saturate(frames / 32.0));
    float sigmaMul  = mix(euler, sqr(0.41), saturate(frames / 16.0));

    float sigmaDistMul = 2.0 - (1.0 / (1.0 + normalDepthCenter.a / 64.0));

	float sigmaL = 1.0 / (sigmaMul * SVGF_STRICTNESS * sigmaDistMul * computeSigmaL(colortex7, pixelUV, colorCenterAux.z) + sigmaBias * sigmaDistMul);

	SH    sumColor    = colorCenter;
	float sumVariance = colorCenterAux.z;
    float sumWeight = 1.0;

    //ivec2 jitter    = getAtrousJitter();

	const int r = SVGF_RAD;
	for(int y = -r; y <= r; ++y) {
		for(int x = -r; x <= r; ++x) {
			ivec2 p = pixelUV + ivec2(x, y) * stepSize;

			if(x == 0 && y == 0)
				continue;

			float w = float(all(greaterThanEqual(p, ivec2(0)))
					&& all(lessThan(p, ivec2(viewSize * indirectResScale))));

            vec4 shAux      = texelFetch(colortex7, p, 0);
			SH col          = SH(texelFetch(colortex6, p, 0), shAux.xy);

			float lum       = getLuma(SH_to_irradiance(col));

			float distLum   = abs(lumaCenter - lum);
                distLum     = sqr(distLum) / max(lumaCenter, offset);
                distLum     = clamp(distLum, 0.0, maxDelta);

            vec4 normalDepth    = texelFetch(imgNormalDepth, p, 0);
            if (normalDepth.a == 1.0) continue;
            normalDepth.rgb = normalDepth.rgb * 2.0 - 1.0;
            normalDepth.a   = sqr(normalDepth.a) * far;

			float distDepth = abs(normalDepthCenter.a - normalDepth.a) * 4.0;

			w *= pow(max(0.0, dot(normalDepthCenter.rgb, normalDepth.rgb)), normalExp);
			w *= exp(-distDepth / sqrt(float(stepSize)) - sqrt(distLum * sigmaL));

            accumulate_SH(sumColor, col, w);

			sumVariance    += shAux.z * sqr(w);
			sumWeight      += w;
		}
	}
    divideSH(sumColor, sumWeight);

	sumVariance /= sqr(sumWeight);

    varOut      = sumVariance;

	return sumColor;
}

void main() {
    resultSH        = texelFetch(colortex6, ivec2(coord * viewSize), 0);
    resultAux       = texelFetch(colortex7, ivec2(coord * viewSize), 0);

    vec2 lowresCoord    = coord / indirectResScale;
    if (saturate(lowresCoord) != lowresCoord) discard;

    #ifdef SVGF_POSTACCUM
    SH result       = filterImage(colortex10, resultAux.z, coord, SVGF_SIZE);

    resultSH        = result.shY;
    resultAux.xy    = result.CoCg;
    #endif

    resultSH        = clamp16FN(resultSH);
    resultAux       = clamp16FN(resultAux);
}